declare namespace Httpr {

    /**
     * Simple hash map.
     */

    export type Map<V> = {
        [key: string]: V;
    };

    /**
     * Available HTTP methods.
     */

    export type HttpMethod = 'get' | 'post' | 'put' | 'delete' | 'options' | 'patch';

    /**
     * One-level depth objects with simple values.
     */

    export type PlainObject = {
        [key: string]: string | number | boolean;
    };

    export class HttpMethods {
        static GET: HttpMethod;
        static POST: HttpMethod;
        static PUT: HttpMethod;
        static DELETE: HttpMethod;
        static OPTIONS: HttpMethod;
        static PATCH: HttpMethod;
    }

    export type HttpRequestSettings = {

        /**
         * Request method.
         */
        method: HttpMethod;

        /**
         * Resource url.
         */
        url: string;

        /**
         * Query parameters.
         */
        params: PlainObject;

        /**
         * Request headers.
         */
        headers: Map<string>;

        /**
         * Request body.
         */
        body?: any;
    };

    export class HttprInterceptor {

        /**
         * Apply modifications to request settings.
         *
         * @param {HttpRequestSettings} settings - current request settings.
         * @return {HttpRequestSettings} modified request settings.
         */
        beforeRequest(settings: HttpRequestSettings): HttpRequestSettings;

        /**
         * Apply modifications to success response.
         *
         * @param {*} response - success response.
         * @return {*} modified response.
         */
        afterSuccess(response: any): any;

        /**
         * Apply modifications to error response.
         *
         * @param {*} response - error response.
         * @return {*} modified response.
         */
        afterError(response: any): any;
    }

    export class HttprProvider {

        /**
         * Perform a request with low level implementation.
         *
         * @param {HttpRequestSettings} settings - request settings.
         * @return {Promise<*>} promise handler.
         */
        request(settings: HttpRequestSettings): Promise<any>;
    }

    export type HttprConfig = {

        /**
         * A base URL to append to all requests.
         */
        baseUrl?: string;

        /**
         * Implementation of the provider that will perform the requests.
         */
        provider: HttprProvider;
    };

    export class HttpHeaders {
        static ACCEPT: string;
        static CONTENT_TYPE: string;
    }

    export class MediaTypes {
        static APPLICATION_JSON: string;
        static FORM_URLENCODED: string;
    }

    /**
     * Encode plain object into query string.
     *
     * @param {PlainObject} object - hash of items.
     * @returns {string} encoded query string.
     */
    export function urlEncode(object: PlainObject): string;

    /**
     * Join two URL parts.
     *
     * @param {Array<string>} parts - URL parts.
     * @return {string} joined URL.
     */
    export function urlJoin(...parts: Array<string>): string;

    export class HttprStatic {

        /**
         * Prepare request settings and apply interceptors before sending the request.
         *
         * @param {Httpr} instance - instance of Http
         * @param {HttpMethod} method - http method to use.
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {Map<string>} headers - hash of request headers to set.
         * @param {*} body - request body data.
         * @returns {HttpRequestSettings} prepared request settings.
         */
        static build(instance: Httpr, method: HttpMethod, url: string, params?: PlainObject, headers?: Map<string>, body?: any): HttpRequestSettings;

        /**
         * Receive success response and apply interceptors.
         *
         * @returns {Promise<*>} resolution handler.
         */
        static onSuccess(instance: Httpr, response: any): Promise<any>;

        /**
         * Receive error response and apply interceptors.
         *
         * @returns {Promise<*>} rejection handler.
         */
        static onError(instance: Httpr, response: any): Promise<any>;
    }

    export class Httpr {

        /**
         * List of interceptors.
         */
        protected _interceptors: Array<HttprInterceptor>;

        /**
         * Httpr configuration.
         */
        protected _config: HttprConfig;

        /**
         * Httpr constructor.
         *
         * @param {HttprConfig} config - custom http configuration.
         */
        constructor(config: HttprConfig);

        /**
         * Return interceptors from this http instance.
         *
         * @returns {Array<HttprInterceptor>} list of interceptors.
         */
        readonly interceptors: Array<HttprInterceptor>;

        /**
         * Return a copy of current configuration.
         *
         * @returns {HttprConfig} current config.
         */
        readonly config: HttprConfig;

        /**
         * Perform a GET request.
         *
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {Map<string>} headers - hash of request headers to set.
         * @returns {Promise<*>}
         */
        get(url: string, params?: PlainObject, headers?: Map<string>): Promise<any>;

        /**
         * Perform a POST request.
         *
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {*} body - request body data.
         * @param {Map<string>} headers - hash of request headers to set.
         * @returns {Promise<*>} promise handler.
         */
        post(url: string, params?: PlainObject, body?: any, headers?: Map<string>): Promise<any>;

        /**
         * Perform a PUT request.
         *
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {*} body - request body data.
         * @param {Map<string>} headers - hash of request headers to set.
         * @returns {Promise<*>} promise handler.
         */
        put(url: string, params?: PlainObject, body?: any, headers?: Map<string>): Promise<any>;

        /**
         * Perform a DELETE request.
         *
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {Map<string>} headers - hash of request headers to set.
         * @returns {Promise<*>} promise handler.
         */
        del(url: string, params?: PlainObject, headers?: Map<string>): Promise<any>;

        /**
         * Perform an OPTIONS request.
         *
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {Map<string>} headers - hash of request headers to set.
         * @returns {Promise<*>} promise handler.
         */
        options(url: string, params?: PlainObject, headers?: Map<string>): Promise<any>;

        /**
         * Perform a PATCH request.
         *
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {Map<string>} headers - hash of request headers to set.
         * @returns {Promise<*>} promise handler.
         */
        patch(url: string, params?: PlainObject, headers?: Map<string>): Promise<any>;

        /**
         * Perform an HTTP Request.
         *
         * @param {HttpMethod} method - http method to use.
         * @param {string} url - url of resource to request.
         * @param {PlainObject} params - hash of additional query parameters.
         * @param {Map<string>} headers - hash of request headers to set.
         * @param {*} body - request body data.
         * @returns {Promise<*>} promise handler.
         */
        request(method: HttpMethod, url: string, params?: PlainObject, headers?: Map<string>, body?: any): Promise<any>;

        /**
         * Add an interceptor to the Http instance.
         *
         * @param {HttprInterceptor} interceptor - instance of the interceptor to add.
         * @returns {Httpr} current Http instance.
         */
        register(interceptor: HttprInterceptor): Httpr;

        /**
         * Remove an interceptor from the Http instance.
         *
         * @param {?HttprInterceptor} interceptor - instance of the interceptor to remove.
         * If none is specified, remove all interceptors.
         * @returns {Httpr} current Http instance.
         */
        unregister(interceptor?: HttprInterceptor): Httpr;
    }
}

declare module 'httpr' { export = Httpr; }